#!/usr/bin/env python
#    Copyright 2014, 2015 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2014 Andrei Macavei <andrei.macavei89@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging
import signal
import sys

import gi
gi.require_version('Gst', '1.0')
gi.require_version('GstVideo', '1.0')
gi.require_version('Gtk', '4.0')
gi.require_version('Gdk', '4.0')
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import Gtk, GLib
# Because of https://bugzilla.gnome.org/show_bug.cgi?id=698005
from gi.repository import Gtk, GdkPixbuf
from gi.repository import GstVideo
from gi.repository import Gdk

log = logging.getLogger(__name__)



class BarcodeReaderGTK(Gtk.Box):

    __gsignals__ = {
        str('barcode'): (GObject.SignalFlags.RUN_LAST, None,
                        (str, # The barcode string
                         Gst.Message.__gtype__, # The GStreamer message itself
                         GdkPixbuf.Pixbuf.__gtype__, # The pixbuf which caused
                                              # the above string to be decoded
                    ),
                   )
    }


    def __init__(self, *args, device=None, pipewire_fd=None, **kwargs):
        super(BarcodeReaderGTK, self).__init__(*args, **kwargs)
        self.device = device
        self.pipewire_fd = pipewire_fd
        self.connect('unmap', self.on_unmap)
        self.connect('map', self.on_map)
        self.scaling_image = ScalingImage()
        self.append(self.scaling_image)


    def on_message(self, bus, message):
        #log.debug("Message: %s", message)
        if message:
            struct = message.get_structure()
            if struct:
                struct_name = struct.get_name()
                #log.debug('Message name: %s', struct_name)

                if struct_name == 'GstMessageError':
                    err, debug = message.parse_error()
                    log.error('GstError: %s, %s', err, debug)
                elif struct_name == 'GstMessageWarning':
                    err, debug = message.parse_warning()
                    log.warning('GstWarning: %s, %s', err, debug)
    
                elif struct_name == 'barcode':
                    self.timestamp = struct.get_clock_time("timestamp")[1]
                    log.debug ("at %s", self.timestamp)

                    assert struct.has_field('symbol')
                    barcode = struct.get_string('symbol')
                    log.info("Read Barcode: {}".format(barcode))

                    pixbuf = None
                    if struct.has_field ("frame"):
                        # This is the new zbar, which posts the frame along
                        # with the barcode.
                        sample = struct.get_value ("frame")
                        pixbuf = gst_sample_to_pixbuf(sample)
                        self.emit("barcode", barcode, message, pixbuf)
                    else:
                        # If we do not see the zbar < 1.6, we raise
                        raise Exception("Zbar version is not what we expected")


    def run(self):
        if self.pipewire_fd is not None:
            src = f"pipewiresrc fd={self.pipewire_fd}"
        elif self.device:
            src = f"v4l2src device={self.device}"
        else:
            src = "autovideosrc"
        pipeline_str = (
            f"{src} "
            " ! videoconvert "
            " ! zbar cache=true attach_frame=true "
            " ! videoconvert "
            " ! appsink sync=false name=imagesink emit-signals=true max-buffers=1 drop=true caps=\"video/x-raw,format=RGBA\""
        )
        log.info("Launching pipeline: %s", pipeline_str)
        pipeline = Gst.parse_launch(pipeline_str)

        self.imagesink = pipeline.get_by_name('imagesink')
        self.imagesink.connect("new-sample", self.on_new_sample)

        self.pipeline = pipeline

        bus = pipeline.get_bus()
        bus.connect('message', self.on_message)
        bus.add_signal_watch()

        pipeline.set_state(Gst.State.PLAYING)


    def set_pipewire_fd(self, fd):
        """Set PipeWire fd for portal-based camera access."""
        log.info("Setting PipeWire fd to: %s", fd)
        was_playing = False
        if hasattr(self, 'pipeline') and self.pipeline:
            state = self.pipeline.get_state(0)[1]
            if state in (Gst.State.PLAYING, Gst.State.PAUSED):
                was_playing = True
            self.pipeline.set_state(Gst.State.NULL)
            self.pipeline = None
        self.pipewire_fd = fd
        self.device = None
        if was_playing:
            self.run()

    def set_device(self, device):
        log.info("Setting device to: %s", device)
        if self.device == device:
            return
        
        is_playing = False
        if hasattr(self, 'pipeline') and self.pipeline:
            state = self.pipeline.get_state(0)[1]
            if state in (Gst.State.PLAYING, Gst.State.PAUSED):
                is_playing = True
            self.pipeline.set_state(Gst.State.NULL)
            self.pipeline = None
            
        self.device = device
        
        if is_playing:
            self.run()


    def on_new_sample(self, appsink):
        sample = appsink.emit("pull-sample")
        if not sample:
            return Gst.FlowReturn.ERROR

        buf = sample.get_buffer()
        caps = sample.get_caps()
        struct = caps.get_structure(0)
        width = struct.get_value("width")
        height = struct.get_value("height")

        data = buf.extract_dup(0, buf.get_size())
        gbytes = GLib.Bytes.new_take(data)

        pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
            gbytes,
            GdkPixbuf.Colorspace.RGB,
            True,
            8,
            width,
            height,
            width * 4
        )

        GLib.idle_add(self.scaling_image.set_from_pixbuf, pixbuf)
        return Gst.FlowReturn.OK


    def pause(self):
        self.pipeline.set_state(Gst.State.PAUSED)


    def on_map(self, *args, **kwargs):
        '''It seems this is called when the widget is becoming visible'''
        self.run()


    def on_unmap(self, *args, **kwargs):
        '''Hopefully called when this widget is hidden,
        e.g. when the tab of a notebook has changed'''
        self.pipeline.set_state(Gst.State.PAUSED)
        # Actually, we stop the thing for real
        self.pipeline.set_state(Gst.State.NULL)


    def do_barcode(self, barcode, message, image):
        "This is called by GObject, I think"
        log.debug("Emitting a barcode signal %s, %s, %r",
                  barcode, message, image)





class ReaderApp(Gtk.Application):
    '''A simple application for scanning a bar code
    
    It makes use of the BarcodeReaderGTK class and connects to
    its on_barcode signal.
    
    You need to have called Gst.init() before creating a
    BarcodeReaderGTK.
    '''
    def __init__(self, *args, **kwargs):
        super(ReaderApp, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)

    
    def on_activate(self, data=None):
        window = Gtk.ApplicationWindow(application=self)
        window.set_title("Gtk Gst Barcode Reader")
        reader = BarcodeReaderGTK()
        reader.connect('barcode', self.on_barcode)
        window.set_child(reader)

        window.present()


    def on_barcode(self, reader, barcode, message, image):
        '''All we do is logging the decoded barcode'''
        logging.info('Barcode decoded: %s', barcode)



class SimpleInterface(ReaderApp):
    '''We tweak the UI of the demo ReaderApp a little'''
    def on_activate(self, *args, **kwargs):
        window = Gtk.ApplicationWindow(application=self)
        window.set_title("Simple Barcode Reader")
        window.set_default_size(400, 300)

        vbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        vbox.set_margin_top(3)
        vbox.set_margin_bottom(3)
        window.set_child(vbox)

        reader = BarcodeReaderGTK()
        reader.connect('barcode', self.on_barcode)
        vbox.append(reader)
        self.reader = reader

        #self.image = Gtk.Image()
        # FIXME: We could show a default image like "no barcode scanned just yet"
        self.image = ScalingImage()
        self.imagebox = Gtk.Box() #expand=True)
        self.imagebox.append(self.image)
        vbox.append(self.imagebox)


        self.playButton = Gtk.ToggleButton()
        self.playButton.set_icon_name("media-playback-pause-symbolic")
        self.playButton.connect("toggled", self.playToggled)
        vbox.append(self.playButton)

        window.present()


    def playToggled(self, button):
        if button.get_active():
            self.reader.pause()
            button.set_icon_name("media-playback-start-symbolic")
        else:
            self.reader.pipeline.set_state(Gst.State.PLAYING)
            button.set_icon_name("media-playback-pause-symbolic")


    def on_barcode(self, reader, barcode, message, pixbuf):
        log.info("Barcode!!1 %r", barcode)

        # Hrm. Somehow, the Gst Widget is allocating
        # space relatively aggressively.  Our imagebox on
        # the right side does not get any space.
        #self.imagebox.remove(self.image)
        #self.image = ScalingImage(pixbuf)
        #self.imagebox.pack_start(self.image, True, True, 0)
        #self.image.set_property('expand', True)
        #self.image.show()
        self.image.set_from_pixbuf(pixbuf)

        # So we just show a window instead...
        w = Gtk.Window()
        w.set_child(ScalingImage(pixbuf))
        w.present()
        return False



def gst_sample_to_pixbuf(sample):
    '''Converts the image from a given GstSample to a GdkPixbuf'''
    caps = Gst.Caps.from_string("video/x-raw,format=RGBA")
    converted_sample = GstVideo.video_convert_sample(sample, caps, Gst.CLOCK_TIME_NONE)

    buffer = converted_sample.get_buffer()
    pixbuf = buffer.extract_dup(0, buffer.get_size())
    
    caps = converted_sample.get_caps()
    struct = caps.get_structure(0)
    
    colorspace = GdkPixbuf.Colorspace.RGB
    alpha = True
    bps = 8
    width_struct = struct.get_int("width")
    assert width_struct[0]
    height_struct = struct.get_int("height")
    assert height_struct[0]
    original_width = width_struct[1]
    original_height = height_struct[1]

    rowstride_struct = struct.get_int("stride")
    if rowstride_struct[0] == True:
        # The stride information might be hidden in the struct.
        # For now it doesn't work. I think it's the name of the field.
        rowstride = rowstride_struct[1]
    else:
        rowstride = bps / 8 * 4 * original_width

    gdkpixbuf = GdkPixbuf.Pixbuf.new_from_bytes(
        GLib.Bytes.new_take(pixbuf),
        colorspace, alpha, bps, original_width,
        original_height, rowstride)
        
    return gdkpixbuf




class ScalingImage(Gtk.DrawingArea):

    def __init__(self, pixbuf=None, width=None, height=None, rowstride=None):
        self.pixbuf = pixbuf
        self.rowstride = rowstride or None
        super(ScalingImage, self).__init__()
        #self.set_property("width_request", 400)
        #self.set_property("height_request", 400)
        #self.set_property("margin", 10)
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.set_draw_func(self.draw_func, None)
    
    def draw_func(self, drawing_area, cr, width, height, user_data):
        self.do_draw(cr, widget_width=width, widget_height=height)
    
    def set_from_pixbuf(self, pixbuf):
        self.pixbuf = pixbuf
        self.queue_draw()


#    def do_size_allocate(self, allocation):
#        log.debug("Size Allocate  %r", allocation)
#        log.debug("w: %r  h: %r",  allocation.width, allocation.height)
#        self.queue_draw()

    def do_draw(self, cr, pixbuf=None, widget_width=None, widget_height=None):
        log.debug('Drawing ScalingImage! %r', self)
        pixbuf = pixbuf or self.pixbuf
        if not pixbuf:
            log.info('No pixbuf to draw! %r', pixbuf)
        else:
            original_width = pixbuf.get_width()
            original_height = pixbuf.get_height()
    
            assert original_width > 0
            assert original_height > 0
    
    
            # Scale the pixbuf down to whatever space we have
            if widget_width is None or widget_height is None:
                allocation = self.get_allocation()
                widget_width = allocation.width
                widget_height = allocation.height
            
            
            # I think we might not need this calculation
            #widget_size = min(widget_width, widget_height)
            log.info('Allocated size: %s, %s', widget_width, widget_height)
            
            # Fill in background
            cr.save()
            #Gtk.render_background(self.get_style_context(),
            #       cr, 0, 0, widget_width, widget_height)
            #cr.set_source_rgb(1, 1, 1)
            #cr.paint()
            
            # Centering and scaling the image to fit the widget
            cr.translate(widget_width / 2.0, widget_height / 2.0)
            scale = min(widget_width / float(original_width), widget_height / float(original_height))
            cr.scale(scale, scale)
            
            cr.translate(-original_width / 2.0, -original_height / 2.0)
            # Note: This function is very inefficient
            # (one could cache the resulting pattern or image surface)!
            Gdk.cairo_set_source_pixbuf(cr, pixbuf, 0, 0)
            # Should anyone want to set filters, this is the way to do it.
            #pattern = cr.get_source()
            #pattern.set_filter(cairo.FILTER_NEAREST)
            cr.paint()
            cr.restore()
            
            return
            #super(ScalingImage, self).do_draw(cr)



def main():
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
                        format='%(name)s (%(levelname)s): %(message)s')

    # We need to have GStreamer initialised before creating a BarcodeReader
    Gst.init(sys.argv)
    app = SimpleInterface()

    try:
        # Exit the mainloop if Ctrl+C is pressed in the terminal.
        GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT, lambda *args : app.quit(), None)
    except AttributeError:
        # Whatever, it is only to enable Ctrl+C anyways
        pass

    app.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import logging
from pathlib import Path
import os
import sys

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .gpgmeh import export_uids, minimise_key

def escape_filename(fname):
    escaped = ''.join(c if c.isalnum() else "_" for c in fname)
    return escaped

def main():
    fname = sys.argv[1]
    keydata = open(fname, 'rb').read()

    minimise = True
    if minimise:
        keydata = minimise_key(keydata)

    for i, (uid, uid_bytes) in enumerate(export_uids(keydata), start=1):
        uid_file = Path('.') / ("{:02d}-".format(i) + escape_filename(uid) + ".pgp.asc")
        print (f"Writing {uid_file}...")
        uid_file.write_bytes(uid_bytes)

    print (f"Done!")

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# encoding: utf-8
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import signal
import sys
import argparse
import logging
import os

import gi
gi.require_version('Gtk', '4.0')
from gi.repository import Gtk, GLib

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .__init__ import __version__
from .gpgmeh import get_usable_keys
from .QRCode import QRImage
from .util import format_fingerprint


log = logging.getLogger(__name__)



class KeyPresentWidget(Gtk.Widget):
    """A widget for presenting a gpgmeh.Key

    It shows details of the given key and customizable data in a
    qrcode encoded in a QRImage.

    The widget takes a full key object, rather than a fingerprint,
    because it makes assumptions about the key object anyway. So
    it can as well take it directly and enable higher level controllers
    to deal with the situation that a given fingerprint, which really is
    a search string for gpg, yields multiple results.
    """

    def __new__(cls, *args, **kwargs):
        thisdir = os.path.dirname(os.path.abspath(__file__))
        builder = kwargs.pop("builder", None)
        if not builder:
            builder = Gtk.Builder()
            builder.add_objects_from_file(
                os.path.join(thisdir, 'send4.ui'),
                ['box3']
                )
        log.debug("Our builder is: %r", builder)
        # The widget will very likely have a parent.
        # Gtk doesn't like so much adding a Widget to a container
        # when the widget already has been added somewhere.
        # So we get the parent widget and remove the child.
        w = builder.get_object('box3')
        parent = w.get_parent()
        if parent:
            parent.remove(w)
        w._builder = builder
        w.__class__ = cls
        return w

    def __init__(self, key, discovery_code, qrcodedata=None, builder=None):
        """
        A new KeyPresentWidget shows the string you provide as qrcodedata
        in a qrcode. If it evaluates to False, the key's fingerprint will
        be shown. That is, "OPENPGP4FPR: + fingerprint.

        :param key: a QR code will be generated using this key's fingerprint if qrcodedata is None
        :param discovery_code: the code that is displayed to the user in the send tab.
        When a user wants to receive a key he can type this code manually instead of scanning the QR code.
        :param qrcodedata: this string will be used in the QR code.
        If None, the key's fingerprint will be used.
        :param builder: not used
        """

        self.key_id_label = self._builder.get_object("keyidLabel")
        self.uids_label = self._builder.get_object("uidsLabel")
        self.fingerprint_label = self._builder.get_object("keyFingerprintLabel")
        self.qrcode_frame = self._builder.get_object("qrcode_frame")

        self.key_id_label.set_markup(
            format_fingerprint(key.fingerprint).replace('\n', '  '))
        self.uids_label.set_markup("\n".join(
                                        [GLib.markup_escape_text(uid.uid)
                                        for uid
                                        in key.uidslist]))
        self.fingerprint_label.set_markup(format_fingerprint(key.fingerprint))
        self.key_fingerprint = key.fingerprint

        self.fingerprint_label.set_markup(discovery_code)

        if not qrcodedata:
            qrcodedata = "OPENPGP4FPR:" + self.key_fingerprint
        self.qrcode_frame.set_child(QRImage(qrcodedata))


class KeyPresent(Gtk.Application):
    """A demo application showing how to display sufficient details
    about a key such that it can be sent securely.
    
    Note that the main purpose is to enable secure transfer, not
    reviewing key details.  As such, the implementation might change
    a lot, depending on the method of secure transfer.
    """
    def __init__(self, *args, **kwargs):
        #super(Keys, self).__init__(*args, **kwargs)
        Gtk.Application.__init__(
            self, application_id="org.gnome.keysign.keypresent")
        self.connect("activate", self.on_activate)
        self.connect("startup", self.on_startup)

        self.log = logging.getLogger(__name__)

        self.key_present_page = None


    def on_quit(self, app, param=None):
        self.quit()


    def on_startup(self, app):
        self.log.info("Startup")
        self.window = Gtk.ApplicationWindow(application=app)
        self.window.set_title ("Keysign - Key")
        self.window.set_child(self.key_present_page)


    def on_activate(self, app):
        self.log.info("Activate!")
        #self.window = Gtk.ApplicationWindow(application=app)

        # In case the user runs the application a second time,
        # we raise the existing window.
        self.window.present()


    def run(self, args):
        log.debug("running: %s", args)
        fpr = args

        key = next(iter(get_usable_keys(pattern=fpr)))
        self.key_present_page = KeyPresentWidget(key, format_fingerprint(key.fingerprint))

        super(KeyPresent, self).run()


def parse_command_line(argv):
    """Parse command line argument. See -h option

    :param argv: arguments on the command line must include caller file name.
    """
    formatter_class = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description='Auxiliary helper program '+
                                                 'to present a key',
                                     formatter_class=formatter_class)
    parser.add_argument("--version", action="version",
                        version="%(prog)s {}".format(__version__))
    parser.add_argument("-v", "--verbose", dest="verbose_count",
                        action="count", default=0,
                        help="increases log verbosity for each occurrence.")
    #parser.add_argument("-g", "--gpg",
    #                    action="store_true", default=False,
    #                    help="Use local GnuPG Keyring instead of file.")
    #parser.add_argument('-o', metavar="output",
    #                    type=argparse.FileType('w'), default=sys.stdout,
    #                    help="redirect output to a file")
    #parser.add_argument('file', help='File to read keydata from ' +
    #                                 '(or KeyID if --gpg is given)')
    parser.add_argument('fpr', help='The fingerprint of the key to transfer')
    ## nargs='+', # argparse.REMAINDER,
    #parser.add_argument('input', metavar="input",
    ## nargs='+', # argparse.REMAINDER,
    #help="input if any...")
    arguments = parser.parse_args(argv[1:])
    # Sets log level to WARN going more verbose for each new -v.
    log.setLevel(max(3 - arguments.verbose_count, 0) * 10)
    return arguments


def main(args=sys.argv):
    """This is an example program of how to use the PresentKey widget"""
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
                        format='%(name)s (%(levelname)s): %(message)s')
    try:
        arguments = parse_command_line(args)
        
        #if arguments.gpg:
        #    keydata = export_keydata(next(get_usable_keys(keyid)))
        #else:
        #    keydata = open(arguments.file, 'r').read()
        fpr = arguments.fpr

        app = KeyPresent()
        try:
            GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT, lambda *args : app.quit(), None)
        except AttributeError:
            pass
    
        exit_status = app.run(fpr)
    
        return exit_status

        
    finally:
        logging.shutdown()

if __name__ == "__main__":
    sys.exit(main())



import logging
from twisted.internet.defer import inlineCallbacks

from .wormholeoffer import WormholeOffer
from .avahioffer import AvahiHTTPOffer
try:
    from .bluetoothoffer import BluetoothOffer
except ImportError:
    BluetoothOffer = None

log = logging.getLogger(__name__)


class Offer:
    def __init__(self, key, app_id=None, w_code=None):
        self.key = key
        self.app_id = app_id
        self.w_code = w_code
        self.w_offer = None
        self.a_offer = None
        self.bt_offer = None
        self.b_data = None

    @inlineCallbacks
    def allocate_code(self, worm=True):
        self.a_offer = AvahiHTTPOffer(self.key)
        code, a_data = self.a_offer.allocate_code()
        discovery_data = [a_data]
        if worm:
            self.w_offer = WormholeOffer(self.key)
            w_info = yield self.w_offer.allocate_code()
            code, w_data = w_info
            if w_data:
                discovery_data.append(w_data)
        if BluetoothOffer:
            self.bt_offer = BluetoothOffer(self.key)
            self.b_data = yield self.bt_offer.allocate_code()
            if self.b_data:
                discovery_data.append(self.b_data)
        discovery_data = ";".join(discovery_data)
        # As design when we use both avahi and wormhole we only display
        # the wormhole code
        return code, discovery_data

    def start(self):
        avahi_defers = self.a_offer.start()
        d = [avahi_defers] if avahi_defers else []
        if self.w_offer:
            w_d = self.w_offer.start()
            d.append(w_d)
        # If we have a Bluetooth code, so if the Bluetooth has been
        # correctly initialized
        if not self.b_data:
            log.info("Bluetooth has been skipped")
        else:
            bt_d = self.bt_offer.start()
            d.append(bt_d)
        return d

    def stop_avahi(self):
        if self.a_offer:
            self.a_offer.stop()
            # We need to deallocate the avahi object or the used port will never be released
            self.a_offer = None

    def stop_wormhole(self):
        if self.w_offer:
            self.w_offer.stop()
            self.w_offer = None

    def stop_bt(self):
        if self.bt_offer:
            self.bt_offer.stop()
            self.bt_offer = None

    def stop(self):
        self.stop_avahi()
        self.stop_wormhole()
        self.stop_bt()



#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2014 Andrei Macavei <andrei.macavei89@gmail.com>
#    Copyright 2015 Jody Hansen <jobediah.hansen@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GObject

if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    __package__ = str('keysign')

from .AvahiConstants import AvahiConstants as avahi

from ..errors import NoAvahiDbus

__all__ = ["AvahiBrowser"]

DBusGMainLoop( set_as_default=True )

log = logging.getLogger(__name__)

# This should probably be upstreamed.
# Unfortunately, upstream seems rather inactive.
if getattr(avahi, 'txt_array_to_dict', None) is None:
    # This has been taken from Gajim
    # https://dev.gajim.org/gajim/gajim/blob/2d6e7d2e/gajim/common/zeroconf/zeroconf_avahi.py#L131
    # it is licensed under the GPLv3.
    # https://github.com/lathiat/avahi/pull/133
    def txt_array_to_dict(txt_array):
        txt_dict = {}
        for els in txt_array:
            key, val = '', None
            for c in els:
                c = chr(c)
                if val is None:
                    if c == '=':
                        val = ''
                    else:
                        key += c
                else:
                    val += c
            if val is None:  # missing '='
                val = ''
            txt_dict[key] = val
        return txt_dict

    setattr(avahi, 'txt_array_to_dict', txt_array_to_dict)


class AvahiBrowser(GObject.GObject):
    __gsignals__ = {
        str('new_service'): (GObject.SignalFlags.RUN_LAST, None,
            # name, address (could be an int too (for IPv4)), port, txt_dict
            (str, str, int, object)),
        str('remove_service'): (GObject.SignalFlags.RUN_LAST, None,
            # string 'remove'(placeholder: tuple element must be sequence), name
            (str, str)),
    }


    def __init__(self, loop=None, service='_gnome-keysign._tcp'):
        GObject.GObject.__init__(self)

        self.log = logging.getLogger(__name__)
        self.service = service
        # It seems that these are different loops..?!
        self.loop = loop or DBusGMainLoop()
        self.bus = dbus.SystemBus(mainloop=self.loop)

        try:
            self.server = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, '/'),
                                         'org.freedesktop.Avahi.Server')
        except dbus.exceptions.DBusException as de:
            log.exception("Avahi cannot be started: %s", de)
            raise NoAvahiDbus(de)

        self.sbrowser = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME,
            self.server.ServiceBrowserNew(avahi.IF_UNSPEC,
                avahi.PROTO_UNSPEC, self.service, 'local', dbus.UInt32(0))),
            avahi.DBUS_INTERFACE_SERVICE_BROWSER)

        self.sbrowser.connect_to_signal("ItemNew", self.on_new_item)
        self.sbrowser.connect_to_signal("ItemRemove", self.on_service_removed)


    def on_new_item(self, interface, protocol, name, stype, domain, flags):
        self.log.info("Found service '%s' type '%s' domain '%s' ", name, stype, domain)

        if flags & avahi.LOOKUP_RESULT_LOCAL:
            # FIXME skip local services
            pass
        self.server.ResolveService(interface, protocol, name, stype,
            domain, avahi.PROTO_UNSPEC, dbus.UInt32(0),
            reply_handler=self.on_service_resolved,
            error_handler=self.on_error)


    def on_service_resolved(self, interface, protocol, name, stype, domain,
                                  host, aprotocol, address, port, txt, flags):
        '''called when the browser successfully found a service'''
        txt = avahi.txt_array_to_dict(txt)
        self.log.info("Service resolved; name: '%s', address: '%s',"
                "port: '%s', and txt: '%s'", name, address, port, txt)
        retval = self.emit('new_service', name, address, port, txt)
        self.log.info("emitted '%s'", retval)


    def on_service_removed(self, interface, protocol, name, stype, domain, flags):
        '''Emits items to be removed from list of discovered services.'''
        self.log.info("Service removed; name: '%s'", name)
        retval = self.emit('remove_service', 'remove', name)
        self.log.info("emitted '%s'", retval)


    def on_error(self, *args):
        print('error_handler')
        print(args[0])


def main():
    loop = GObject.MainLoop()
    # We're not passing the loop to DBus, because... well, it
    # doesn't work... It seems to expect a DBusMainLoop, not
    # an ordinary main loop...
    ab = AvahiBrowser()

    def print_signal(*args):
        print("Signal ahoi", args)

    ab.connect('new_service', print_signal)
    ab.connect('remove_service', print_signal)
    loop.run()


if __name__ == "__main__":
    main()



class BluetoothException(Exception):
    """Parent class for all Bluetooth-related errors"""


class NoBluezDbus(BluetoothException):
    """The required org.bluez dbus is not available"""


class NoAdapter(BluetoothException):
    """There isn't an usable Bluetooth adapter"""


class UnpoweredAdapter(BluetoothException):
    """The Bluetooth adapter is turned off"""


class AvahiException(Exception):
    """Parent class for all Avahi-related errors"""


class NoAvahiDbus(AvahiException):
    """The required avahi dbus is not available"""



#!/usr/bin/env python
# encoding: utf-8
#    Copyright 2016 Andrei Macavei <andrei.macavei89@gmail.com>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import sys
import logging
import os

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Gst', '1.0')

from gi.repository import Gtk, Gst, GdkPixbuf
from gi.repository import GObject


if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .scan_barcode import BarcodeReaderGTK
from . import camera_portal

log = logging.getLogger(__name__)



class KeyFprScanWidget(Gtk.Box):
    """A widget for obtaining a key fingerprint.

    The fingerprint can be obtain by inserting it into
    a text entry, or by scanning a barcode with the
    built-in camera.
    """

    __gsignals__ = {
        # This is the Gtk widget signal's name
        str('changed'): (GObject.SignalFlags.RUN_LAST, None,
                        (GObject.TYPE_PYOBJECT,)),
        # It's probably not the best name for that signal.
        # While "barcode_scanned" might be better, it is probably
        # unnecessarily specific.
        str('barcode'): (GObject.SignalFlags.RUN_LAST, None,
                        (str, # The barcode string
                         Gst.Message.__gtype__, # The GStreamer message itself
                         GdkPixbuf.Pixbuf.__gtype__,),) # The pixbuf which caused
                                              # the above string to be decoded
    }

    def __init__(self, builder=None):
        log.debug("Init KFSW %r %r", self, builder)
        if issubclass(self.__class__, object):
            super(KeyFprScanWidget, self).__init__(orientation=Gtk.Orientation.VERTICAL)
        else:
            Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        log.debug("Inited parent KFSW %r", self)

        widget_name = 'scanner_widget'
        if not builder:
            thisdir = os.path.dirname(os.path.abspath(__file__))
            builder = Gtk.Builder()
            builder.add_objects_from_file(os.path.join(thisdir, 'receive4.ui'),
                [widget_name])
        widget = builder.get_object(widget_name)
        parent = widget.get_parent()
        if parent:
            parent.remove(widget)
        self.append(widget)
        

        self.scanner = builder.get_object("scanner")

        if not Gst.is_initialized():
            log.error("Gst does not seem to be initialised. Call Gst.init()!")
            # This needs to be called before creating a BarcodeReaderGTK
            Gst.init(None)
        reader = BarcodeReaderGTK()
        reader.set_size_request(150,150)
        reader.connect('barcode', self.on_barcode)
        self.scanner.append(reader)
        # We keep a reference here to not "lose" the object.
        # If we don't, Gtk crashes. With a segfault. Probably
        # because the object is freed but still used.
        # Somebody should look at that...
        self.reader = reader

        self.camera_selector = builder.get_object("comboboxtext1")
        self.camera_box = builder.get_object("box40")
        self.camera_devices = {}
        
        self._using_portal = False
        if camera_portal._using_flatpak() and camera_portal.is_camera_portal_available():
            log.info("Running in Flatpak and Camera Portal is available, requesting access")
            self._using_portal = True
            # Hide the camera selector dropdown — portal handles camera access
            if self.camera_box:
                self.camera_box.set_visible(False)
            camera_portal.request_camera_access(self._on_camera_portal_response)
        else:
            # Legacy path: enumerate devices with Gst.DeviceMonitor
            if self.camera_selector:
                self.populate_cameras()

    def _on_camera_portal_response(self, success, pipewire_fd):
        """Called when the Camera Portal responds to our access request."""
        if success and pipewire_fd is not None:
            log.info("Camera Portal granted access, fd=%d", pipewire_fd)
            from gi.repository import GLib
            GLib.idle_add(self.reader.set_pipewire_fd, pipewire_fd)
        else:
            log.warning("Camera Portal denied or failed, falling back to "
                        "direct device access.")
            self._using_portal = False
            from gi.repository import GLib
            GLib.idle_add(self._fallback_to_device_monitor)

    def _fallback_to_device_monitor(self):
        """Fall back to legacy Gst.DeviceMonitor enumeration."""
        if self.camera_box:
            self.camera_box.set_visible(True)
        if self.camera_selector:
            self.populate_cameras()

        self.fpr_entry = builder.get_object("fingerprint_entry")
        self.fpr_entry.connect('changed', self.on_text_changed)
        
        self.set_hexpand(True)
        self.set_vexpand(True)

        # Temporary measure...
        self.barcode_scanner = self

    def populate_cameras(self):
        self.camera_selector.remove_all()
        
        monitor = Gst.DeviceMonitor.new()
        monitor.add_filter("Video/Source", None)
        monitor.start()
        devices = monitor.get_devices()
        monitor.stop()
        
        self.camera_devices = {}
        best_suitable_idx = -1
        best_suitable_v4l2 = -1
        
        best_unsuitable_idx = -1
        best_unsuitable_v4l2 = -1
        
        import re
        def get_v4l2_index(path):
            if not path:
                return -1
            m = re.search(r'\d+$', path)
            return int(m.group(0)) if m else -1
        
        for idx, device in enumerate(devices):
            display_name = device.get_display_name()
            props = device.get_properties()
            device_path = None
            if props:
                for key in ["api.v4l2.path", "device.path", "object.path"]:
                    val = props.get_string(key)
                    if val:
                        if val.startswith("v4l2:"):
                            val = val[5:]
                        device_path = val
                        break
            
            if not device_path:
                continue
                
            name_lower = display_name.lower()
            is_unsuitable = "ir" in name_lower or "infrared" in name_lower or "infra-red" in name_lower
            
            v4l2_num = get_v4l2_index(device_path)
            current_idx = len(self.camera_devices)
            item_id = str(current_idx)
            self.camera_devices[item_id] = device_path
            
            if is_unsuitable:
                label = f"⚠️ {display_name} ({device_path}) [IR / Unsuitable]"
                if v4l2_num > best_unsuitable_v4l2:
                    best_unsuitable_v4l2 = v4l2_num
                    best_unsuitable_idx = current_idx
            else:
                label = f"{display_name} ({device_path})"
                if v4l2_num > best_suitable_v4l2:
                    best_suitable_v4l2 = v4l2_num
                    best_suitable_idx = current_idx
            
            self.camera_selector.append(item_id, label)
            
        if self.camera_devices:
            self.camera_selector.connect("changed", self.on_camera_changed)
            if best_suitable_idx != -1:
                default_index = best_suitable_idx
            elif best_unsuitable_idx != -1:
                default_index = best_unsuitable_idx
            else:
                default_index = 0
            self.camera_selector.set_active(default_index)
            default_path = self.camera_devices.get(str(default_index))
            if default_path:
                self.reader.set_device(default_path)

    def on_camera_changed(self, combo):
        active_id = combo.get_active_id()
        if active_id and active_id in self.camera_devices:
            device_path = self.camera_devices[active_id]
            log.info("Camera changed in dropdown to ID %s: %s", active_id, device_path)
            self.reader.set_device(device_path)

    def on_text_changed(self, entryObject, *args):
        self.emit('changed', entryObject, *args)

    def on_barcode(self, sender, barcode, message, image):
        self.emit('barcode', barcode, message, image)

    def get_text(self):
        "Returns the text present in the Entry"
        text = self.fpr_entry.get_text()
        return text

class KeyScanApp(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super(KeyScanApp, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)
        self.scanwidget = None

        self.log = logging.getLogger(__name__)

    def on_activate(self, app):
        window = Gtk.ApplicationWindow()
        window.set_title("Key Fingerprint Scanner Widget")
        window.set_size_request(600, 400)

        if not self.scanwidget:
            self.scanwidget = KeyFprScanWidget()
        self.scanwidget.connect('changed', self.on_text_changed)
        self.scanwidget.connect('barcode', self.on_barcode)
        window.set_child(self.scanwidget)
        window.present()
        self.add_window(window)

    def on_text_changed(self, keyFprScanWidget, entryObject, *args):
        self.log.debug ("Text changed! %s" % (entryObject.get_text(),))

    def on_barcode(self, sender, barcode, message, image):
        self.log.debug ("Barcode signal %r %r", barcode, message)


if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG)
    Gst.init(None)
    app = KeyScanApp()
    app.run()



#!/usr/bin/env python

import gettext
import locale
import logging
import os

log = logging.getLogger(__name__)

APP = 'keysign'
# We have pretty much no idea what we're doing here.
# We want to be able to find the compiled message catalogues
# not only when the app has been installed, but also
# when it's being run from source.  Hence we're using
# a path relative to this file, hoping that it will cater
# for both use cases.  Feel free to improve.
DIR = os.path.join(os.path.dirname(__file__), 'locale')

## This is for C libraries, I think. I.e. not for pure python
## ones like us.  We do, however, need this for Gtk.Builder
## to load translations
try:
    locale.setlocale(locale.LC_ALL, '')
except locale.Error:
    log.exception("Cannot set locale")
locale.bindtextdomain(APP, DIR)
locale.textdomain(APP)
gettext.bindtextdomain(APP, DIR)
gettext.textdomain(APP)
lang = gettext.translation(APP, DIR, fallback=True)

## This seems to cause "_" to be installed into globals.
## Let's not do that to not confuse libraries that we include
# gettext.install(APP, DIR, unicode=1)


from locale import gettext as _

try:
    _ = lang.ugettext
except AttributeError:
    _ = lang.gettext



import logging
import dbus
import select
import socket
import sys

if __name__ == "__main__":
    import gi
    gi.require_version('Gtk', '4.0')
    from twisted.internet import gireactor
    gireactor.install()
    from twisted.internet import reactor
from twisted.internet import threads
from twisted.internet.defer import inlineCallbacks

if sys.version < '3':
    input = raw_input

if __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    import os
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    __package__ = str('keysign')

from .errors import NoBluezDbus, NoAdapter, UnpoweredAdapter
from .gpgmeh import get_public_key_data, get_usable_keys
from .i18n import _
from .util import get_local_bt_address, mac_generate

log = logging.getLogger(__name__)


class BluetoothOffer:
    def __init__(self, key, port=3, size=1024):
        self.key = key
        self.port = port
        self.size = size
        self.server_socket = None
        self.message_def = None
        self.stopped = False

    @inlineCallbacks
    def start(self):
        self.stopped = False
        message = "Back"
        success = False
        try:
            while not self.stopped and not success:
                # server_socket.accept() is not stoppable. So with select we can call accept()
                # only when we are sure that there is already a waiting connection
                ready_to_read, ready_to_write, in_error = yield threads.deferToThread(
                    select.select, [self.server_socket], [], [], 0.5)
                if ready_to_read:
                    # We are sure that a connection is available, so we can call
                    # accept() without deferring it to a thread
                    client_socket, address = self.server_socket.accept()
                    key_data = get_public_key_data(self.key.fingerprint)
                    kd_decoded = key_data.decode('utf-8')
                    yield threads.deferToThread(client_socket.sendall, kd_decoded)
                    log.info("Key has been sent")
                    client_socket.shutdown(socket.SHUT_RDWR)
                    client_socket.close()
                    success = True
                    message = None
        except Exception as e:
            log.error("An error occurred: %s" % e)
            success = False
            message = e

        return success, message

    @inlineCallbacks
    def allocate_code(self):
        """Acquires and returns a string suitable for finding the key via Bluetooth.
        Returns None if no powered on adapter could be found."""
        bt_data = None
        log.debug("BT: Allocating code for %s", self.key)
        try:
            code = yield threads.deferToThread(get_local_bt_address)
            log.debug("local addr: %s", code)
            code = code.upper()
        except NoBluezDbus as e:
            log.debug("Bluetooth service seems to be unavailable: %s", e)
        except NoAdapter as e:
            log.debug("Bluetooth adapter is not available: %s", e)
        except UnpoweredAdapter as e:
            log.debug("Bluetooth adapter is turned off: %s", e)
        else:
            log.info("yo")
            if self.server_socket is None:
                log.info("yo")
                self.server_socket = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_STREAM, socket.BTPROTO_RFCOMM)
                # We create a bind with the Bluetooth address we have in the system
                PORT_ANY = 0
                self.server_socket.bind((code, PORT_ANY))
                # Number of unaccepted connections that the system will allow before refusing new connections
                backlog = 1
                self.server_socket.listen(backlog)
                log.info("sockname: %r", self.server_socket.getsockname())
            port = self.server_socket.getsockname()[1]
            log.info("BT Code: %s %s", code, port)
            bt_data = "BT={0};PT={1}".format(code, port)

        log.info("BT return code: %s", bt_data)
        return bt_data

    def stop(self):
        log.debug("Stopping bt receive")
        self.stopped = True
        if self.server_socket:
            self.server_socket.shutdown(socket.SHUT_RDWR)
            self.server_socket.close()
            self.server_socket = None


def main(args):
    if not args:
        log.info("Usable keys: %s", get_usable_keys())
        print ("Usable keys:")
        for k in get_usable_keys()[:50]:
            print ("%s" % k.fingerprint)
        raise ValueError(_("You must provide an argument to identify the key"))

    def code_generated(data):
        if data:
            # getting the code from "BT=code;...."
            code = data.split("=", 1)[1]
            code = code.split(";", 1)[0]
            port = data.rsplit("=", 1)[1]
            offer.start().addCallback(_received)
            print(_("Offering key: {}").format(key))
            print(_("Discovery info: {}").format(code))
            print(_("HMAC: {}").format(hmac))
            print(_("Port: {}").format(port))
            # Wait for the user without blocking everything
        else:
            print(_("Bluetooth not available"))

        reactor.callInThread(cancel)

    def cancel():
        input(_("Press Enter to cancel"))
        offer.stop()
        reactor.callFromThread(reactor.stop)

    def _received(result):
        success, error_msg = result
        if success:
            print(_("\nKey successfully sent"))
        else:
            print(_("\nAn error occurred: {}").format(error_msg))
        # We are still waiting for the user to press Enter
        print(_("Press Enter to exit"))

    key = get_usable_keys(pattern=args[0])[0]
    file_key_data = get_public_key_data(key.fingerprint)
    hmac = mac_generate(key.fingerprint.encode('ascii'), file_key_data)
    offer = BluetoothOffer(key)

    log.info("Running...")
    offer.allocate_code().addCallback(code_generated)
    reactor.run()


if __name__ == "__main__":
    import sys
    logging.basicConfig(level=logging.DEBUG)
    main(sys.argv[1:])



#!/usr/bin/env python
#    Copyright 2020 Tobias Mueller <tobi@cryptobit.ch>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import argparse

import logging
import sys


from .gpgmeh import sign_keydata_and_encrypt, fingerprint_from_keydata

def main():
    import argparse
    parser = argparse.ArgumentParser(description="Sign an OpenPGP key from a file. "
        "The program will open each file, extract exactly one OpenPGP key, "
        "sign each UID separately, encrypt and write the result out to a file.")
    parser.add_argument('-v', '--verbose', action='count', default=0,
        help="Increase detail of logging")
    parser.add_argument("--preserve", action='store_true',
        help="Write files with plaintext next to the ciphertext, e.g. .plain.pgp")
    parser.add_argument('--sign-and-send-all-uids', action='store_true',
        help="Attach certifications for email-less UIDs to the files generated for UIDs with emails.")
    parser.add_argument("file", nargs='+', type=argparse.FileType('rb'),
        help="File containing OpenPGP keys")
    args = parser.parse_args()

    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = log_levels[min(len(log_levels)-1, args.verbose)]
    logging.basicConfig(level=log_level)

    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)

    for fhandle in args.file:
        data = fhandle.read()
        fingerprint = fingerprint_from_keydata(data)
        signed_uids = list(sign_keydata_and_encrypt(keydata=data))
        
        if args.sign_and_send_all_uids:
            emailable_uids = [(uid, enc, pt) for uid, enc, pt in signed_uids if uid.email and uid.email != 'unknown']
            emailless_uids = [(uid, enc, pt) for uid, enc, pt in signed_uids if not uid.email or uid.email == 'unknown']
            
            emailless_ciphertext = b"".join([enc for _, enc, _ in emailless_uids])
            emailless_plaintext = b"".join([pt for _, _, pt in emailless_uids])
            
            uids_to_process = emailable_uids if emailable_uids else emailless_uids
            
            for i, (uid, ciphertext, plaintext) in enumerate(uids_to_process):
                fname = "%s-%d.pgp" % (fingerprint, i)
                with open(fname, 'wb') as outfile:
                    outfile.write(ciphertext)
                    if emailable_uids and emailless_uids:
                        outfile.write(emailless_ciphertext)
                    print ("Written to %s \t for UID %s" % (fname, uid))
                
                if args.preserve:
                    fname = "%s-%d.plain.pgp" % (fingerprint, i)
                    with open(fname, 'wb') as outfile:
                        outfile.write(plaintext)
                        if emailable_uids and emailless_uids:
                            outfile.write(emailless_plaintext)
                        print ("Written Plaintext to %s \t for UID %s" % (fname, uid))
        else:
            for i, (uid, ciphertext, plaintext) in enumerate(signed_uids):
                fname = "%s-%d.pgp" % (fingerprint, i)
                with open(fname, 'wb') as outfile:
                    outfile.write(ciphertext)
                    print ("Written to %s \t for UID %s" % (fname, uid))
    
                if args.preserve:
                    fname = "%s-%d.plain.pgp" % (fingerprint, i)
                    with open(fname, 'wb') as outfile:
                        outfile.write(plaintext)
                        print ("Written Plaintext to %s \t for UID %s" % (fname, uid))



if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    sys.exit(main())



#!/usr/bin/env python

import logging
import mailbox
import os
import signal
from string import Template
from tempfile import NamedTemporaryFile

try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

import gi
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
from gi.repository import Gtk, Adw
from gi.repository import GLib
from gi.repository import Gdk
from gpg import errors
from wormhole.errors import ServerConnectionError, LonelyError, WrongPasswordError
if __name__ == "__main__":
    from twisted.internet import gireactor
    gireactor.install()
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from .keylistwidget import KeyListWidget
from .KeyPresent import KeyPresentWidget
from .offer import Offer
from .util import get_attachments, send_email
from . import gpgmeh
# We import i18n to have the locale set up for Glade
from .i18n import _
log = logging.getLogger(__name__)

try:
    from .bluetoothoffer import BluetoothOffer
except ImportError:
    log.exception("cannot import BluetoothOffer")
    BluetoothOffer = None


DRAG_ACTION = Gdk.DragAction.COPY



RETURN_SUBJECT = "Your OpenPGP certifications on my key"
RETURN_BODY = """Hi $uid,

thanks for having signed my key.
Here are the certifications you produced.

Please import them by, e.g., spawning a terminal and typing

    gpg --import

Then, drag and drop the attachment into your terminal and press Enter.
"""


class SendApp:
    """Common functionality needed when building the sending part
    
    This class will automatically start the keyserver
    and avahi components.  It will load a GtkStack from "send.ui"
    and automatically switch to a newly generate KeyPresentWidget.
    To switch the stack back and stop the keyserver, you have to
    call deactivate().
    """
    def __init__(self, builder=None):
        self.offer = None
        self.stack = None
        self.stack_saved_visible_child = None
        self.klw = None
        self.kpw = None

        ui_file_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "send4.ui")
        if not builder:
            builder = Gtk.Builder()
            builder.add_objects_from_file(ui_file_path, ["send_stack"])
        keys = gpgmeh.get_usable_secret_keys()
        klw = KeyListWidget(keys, builder=builder)
        klw.connect("key-activated", self.on_key_activated)
        self.klw = klw

        stack = builder.get_object("send_stack")
        if hasattr(stack, 'add_child'):
            stack.add_child(klw)
        else:
            stack.add(klw)
        self.stack = stack

        # This is a dirty hack :-/
        # The problem is that the .ui file contains a few widgets
        # that we (potentially) want to instantiate separately.
        # Now that may not necessarily be what Gtk people envisioned
        # so it's not supported nicely.
        # The actual problem is that the widgets of our desire are
        # currently attached to a GtkStack. When our custom widget
        # code runs, it detaches itself from its parent, i.e. the stack.
        # We need need to instantiate the widget with key, however.
        fakekey = gpgmeh.Key("","","")
        kpw = KeyPresentWidget(fakekey, "", builder=builder)

        self.rb = builder.get_object('resultbox')
        self.stack.remove(self.rb)
        self.key = None
        self.result_label = builder.get_object("result_label")
        self.success_label = builder.get_object("success_label")
        self.password_error_label = builder.get_object("password_error_label")
        self.notify = None
        self.internet_option = False

        self.label = builder.get_object("keys_listbox")

        self.rb.connect('unmap', self.on_resultbox_unmapped)
        self.rb_import_okay = builder.get_object('rb_infobar_import_okay')
        self.rb_button_ib_return_signature = builder.get_object('rb_return_signature')
        self.rb_import_error = builder.get_object('rb_infobar_import_error')
        assert self.rb_import_error
        self.button_rb_import_error = builder.get_object('rb_import_error_details_button')



    def on_drag_data_received(self, widget, drag_context, x, y, data, info, time):
        if self.notify is not None:
            log.debug("We are trying to send a key, no imports at this stage")
            return
        try:
            decrypted_certifications, attestors = self.decrypt_and_import_certifications(data)
        except gpgmeh.NoNewSignatures as e:
            self.no_new_signatures_import_error(e)
        except errors.GPGMEError as e:
            self.signature_import_error(e)
        else:
            self.signature_imported(decrypted_certifications, attestors)

    def on_rb_drag_data_received(self, widget, drag_context, x, y, data, info, time):
        try:
            decrypted_certifications, attestors = self.decrypt_and_import_certifications(data)
        except errors.GPGMEError as e:
            self.rb_signature_import_error(e)
        else:
            self.rb_signature_imported(decrypted_certifications, attestors)


    def decrypt_and_import_certifications(self, data):
        filename = data.get_text()
        # If we don't have a filename it means that the user maybe dropped
        # an attachment or an entire email.
        if not filename:
            filename = data.get_data().decode("utf-8")

        filename = unquote(filename)
        filename = filename[7:].strip('\r\n\x00')  # remove file://, \r\n and NULL
        log.info("Received file: %s", filename)
        signatures = get_attachments(filename)
        if not signatures:
            with open(filename, "rb") as si:
                signatures.append(si.read())

        sigs_before = {key.fingerprint: gpgmeh.get_signatures_for_uids_on_key(key)
                       for key in gpgmeh.get_usable_secret_keys()}
        # We currently do not know how to obtain the sender of the email.
        # We could parse the email for a From: header.
        # But if only the attachment is dropped, we don't have that information.
        # We could probably find the issuer in the hashed area of the certification packet,
        # and then try to find that key in our keyring.
        sender = None

        decrypted_certifications = []
        try:
            for signature in signatures:
                decrypted_certification = gpgmeh.decrypt_signature(signature)
                import_result = gpgmeh.import_signature(decrypted_certification)
                decrypted_certifications.append(decrypted_certification)
        except errors.GPGMEError as e:
            log.exception("Could not import signatures")
            raise
        else:
            sigs_after = {key.fingerprint: gpgmeh.get_signatures_for_uids_on_key(key)
                          for key in gpgmeh.get_usable_secret_keys()}
            log.debug("Delta Sigs, before: %s", sigs_before)
            attestors = set()
            for fpr, uid_sigs in sigs_after.items():
                for uid, sigs in uid_sigs.items():
                    sig_delta = sigs - sigs_before[fpr][uid]
                    log.info("These certifications are new: %s", sig_delta)
                    for keyid in sig_delta:
                        for key in gpgmeh.get_usable_keys(pattern=keyid):
                            log.debug("Attestor key has UIDs: %s", key.uidslist)
                            for uid in key.uidslist:
                                email = uid.email
                                log.debug("Attestor key %s has UID %s with Email %s", key, uid, email)
                                if email:
                                    log.debug("Adding %s to %s", email, attestors)
                                    attestors.add(email)
                                    log.debug("Now, attestors is %s", attestors)

            log.debug("Found attestors: %s", attestors)
            return decrypted_certifications, attestors

    @inlineCallbacks
    def on_key_activated(self, widget, key):
        # Deactivate any old connection attempt
        self._deactivate_timer()
        self.deactivate()
        self.klw.ib_internet.hide()
        self.key = key
        log.info("Activated key %r", key)
        ####
        # Start network services
        self.klw.code_spinner.start()
        if self.internet_option:
            #self.kpw.internet_spinner.start()
            # After 10 seconds without a wormhole code we display an info bar
            timer = 10
            self.notify = reactor.callLater(timer, self.slow_connection)
            self.offer = Offer(self.key)
            try:
                info = yield self.offer.allocate_code(worm=True)
            except ServerConnectionError:
                # We are without a working Internet connection so we stop the previously
                # activated Avahi server and we display an infobar
                self._deactivate_timer()
                self.offer.stop_avahi()
                self.offer = None
                self.klw.code_spinner.stop()
                self.no_connection()
                return

            code, discovery_data = info
            self.create_keypresent(code, discovery_data)
            defers = self.offer.start()
            for de in defers:
                # TODO handle errors here?
                de.addCallback(self._received)
        else:
            self.offer = Offer(self.key)
            info = yield self.offer.allocate_code(worm=False)
            code, discovery_data = info
            self.create_keypresent(code, discovery_data)
            defers = self.offer.start()
            for de in defers:
                de.addCallback(self._received)

    def _received(self, start_data):
        success, message = start_data
        if message and type(message) == LonelyError:
            # This only means that we closed wormhole before a transfer
            pass
        elif message and message == "Back":
            # Simply the return of the back button, no errors here
            pass
        elif message and type(message) == ServerConnectionError:
            self._deactivate_timer()
            self.deactivate()
            self.klw.code_spinner.stop()
            self.no_connection()
        else:
            self.show_result(success, message)

    def on_ib_closed(self, widget, data):
        if Gtk.ResponseType.CLOSE == data:
            widget.hide()

    def slow_connection(self):
        self.klw.label_ib_internet.set_label(_("Still trying to get a connection to the Internet. "
                                             "It appears to be slow or unavailable."))
        self.klw.ib_internet.show()
        log.info("Slow Internet connection")

    def no_connection(self):
        self.klw.label_ib_internet.set_label(_("There isn't an Internet connection!"))
        self.klw.ib_internet.show()
        log.info("No Internet connection")

    def signature_imported(self, decrypted_certifications, attestors):
        """When we have received, decrypted, and imported a certification,
        this function will show the infobar and offer to return the certification
        to the sender.
        We appreciate that the sender, i.e. an email address, is not always known,
        simply because that information is currently not included in the certification
        the attestor sends.
        """
        log.debug("s_i Attestors: %s", attestors)
        self.klw.ib_import_okay.show()
        if not attestors:
            # We do not know where to return the certification to, so we hide the button
            self.klw.button_ib_import_okay.hide()
        else:
            def return_certification(button):
                self._tempfiles = []
                for sender in attestors:
                    log.info("Return certification to %s (%d)",
                        sender, len(decrypted_certifications))

                    tempfiles = []
                    for cert in decrypted_certifications:
                        tempfile = NamedTemporaryFile(prefix='gnome-keysign-certifications',
                                                     suffix='.asc',
                                                     delete=True)
                        tempfile.write(cert)
                        tempfile.file.close()
                        tempfiles.append(tempfile)

                    ctx = {'uid': sender}
                    subject = Template(RETURN_SUBJECT).safe_substitute(ctx)
                    body = Template(RETURN_BODY).safe_substitute(ctx)
                    send_email(sender, subject=subject, body=body, files=[f.name for f in tempfiles])
                    # We just keep the object around so that the files do not get deleted before the email has been sent. Once the app quits, we are happy with the files being cleaned up.
                    self._tempfiles.append(tempfiles)

            self.klw.button_ib_import_okay.connect('clicked', return_certification)
            self.klw.button_ib_import_okay.show()

        log.info("Signature imported")

    def no_new_signatures_import_error(self, e):
        self.klw.ib_import_error_no_new_sigs.show()
        log.info("No new signatures: %r", e)

    def signature_import_error(self, e):
        self.klw.ib_import_error.show()
        # We hide the error details button, because we don't have that functionality just yet
        self.klw.button_ib_import_error.hide()
        log.info("Signature import error: %r", e)

    def rb_signature_imported(self, decrypted_certifications, attestors):
        self.rb_import_okay.show()
        log.debug("rb_s_i Attestors: %s", attestors)
        if not attestors:
            # We do not know where to return the certification to, so we hide the button
            self.rb_button_ib_return_signature.hide()
        else:
            def return_certification(button):
                for sender in attestors:
                    log.info("Return certification to %s (%d)",
                        sender, len(decrypted_certifications))

            self.rb_button_ib_return_signature.connect('clicked', return_certification)
            self.rb_button_ib_return_signature.show()

        log.info("Signature imported")

    def rb_signature_import_error(self, e):
        self.rb_import_error.show()
        # We hide the error details button, because we don't have that functionality just yet
        self.button_rb_import_error.hide()
        log.info("Signature import error")

    def create_keypresent(self, discovery_code, discovery_data):
        self._deactivate_timer()
        log.info("Use this for discovering the other key: %r", discovery_data)
        ####
        # Create widget for key
        self.kpw = KeyPresentWidget(self.key, discovery_code, discovery_data)

        ####
        # Show widget for key
        self.stack.add_child(self.kpw)
        self.stack_saved_visible_child = self.stack.get_visible_child()
        self.stack.set_visible_child(self.kpw)
        log.debug('Setting kpw: %r', self.kpw)
        self.klw.ib_internet.hide()
        self.klw.code_spinner.stop()

    def show_result(self, success, message):
        self._deactivate_offer()

        self.stack.add_child(self.rb)
        self.stack.remove(self.kpw)
        self.kpw = None

        if success:
            self.result_label.hide()
            self.success_label.show()
            self.password_error_label.hide()
            self.stack.set_visible_child(self.rb)
        else:
            if type(message) == WrongPasswordError:
                self.success_label.hide()
                self.result_label.hide()
                self.password_error_label.show()
            else:
                self.success_label.hide()
                self.password_error_label.hide()
                self.result_label.show()
                self.result_label.set_label(_("An unexpected error occurred:\n%s" % message))
            self.stack.set_visible_child(self.rb)

    def deactivate(self):
        self._deactivate_offer()

        ####
        # Re-set stack to initial position
        self.set_saved_child_visible()
        if self.kpw:
            self.stack.remove(self.kpw)
            self.kpw = None

    def set_saved_child_visible(self):
        if self.stack_saved_visible_child:
            self.stack.set_visible_child(self.stack_saved_visible_child)
            self.stack_saved_visible_child = None

    def set_internet_option(self, value):
        self._deactivate_timer()
        self.deactivate()
        self.klw.ib_internet.hide()
        self.klw.code_spinner.stop()
        self.internet_option = value

    def _deactivate_timer(self):
        if self.notify and not self.notify.called:
            self.notify.cancel()
            self.notify = None

    def _deactivate_offer(self):
        # Stop network services
        if self.offer:
            self.offer.stop()
            self.offer = None
            log.debug("Stopped network services")


    def on_resultbox_unmapped(self, rb):
        log.debug("Resultbox disappears %r", rb)



class App(Adw.Application):
    def __init__(self, *args, **kwargs):
        super(App, self).__init__(*args, **kwargs)
        self.connect('activate', self.on_activate)
        self.send_app = None

    def on_activate(self, data=None):
        ui_file_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "send4.ui")
        self.builder = Gtk.Builder.new_from_file(ui_file_path)
        window = self.builder.get_object("appwindow")
        assert window
        window.connect("close-request", self.on_delete_window)
        self.headerbar = self.builder.get_object("headerbar")
        hb = self.builder.get_object("headerbutton")
        hb.connect("clicked", self.on_header_button_clicked)
        self.header_button = hb
        self.internet_toggle = self.builder.get_object("internet_toggle")
        self.internet_toggle.connect("toggled", self.on_toggle_clicked)

        self.send_app = SendApp(builder=self.builder)
        ss = self.send_app.stack
        ss.connect('notify::visible-child', self.on_send_stack_switch)
        ss.connect('map', self.on_send_stack_mapped)
        self.send_stack = ss

        window.present()
        self.add_window(window)

    @staticmethod
    def on_delete_window(*args):
        reactor.callFromThread(reactor.stop)

    def on_toggle_clicked(self, toggle):
        log.info("Internet toggled to: %s", toggle.get_active())
        self.send_app.set_internet_option(toggle.get_active())

    def on_send_stack_switch(self, stack, *args):
        log.debug("Switched Send Stack! %r", args)
        current = self.send_app.stack.get_visible_child()
        if current == self.send_app.klw:
            log.debug("Key list page now visible")
            self.on_keylist_mapped(self.send_app.klw)
        elif current == self.send_app.kpw:
            log.debug("Key present page now visible")
            self.on_keypresent_mapped(self.send_app.kpw)
        elif current == self.send_app.rb:
            log.debug("Result page now visible")
            self.on_resultbox_mapped(self.send_app.rb)
        else:
            log.error("An unexpected page is now visible: %r", current)

    def on_resultbox_mapped(self, rb):
        log.debug("Resultbox becomes visible!")
        self.header_button.set_sensitive(True)
        self.header_button.set_image(
            Gtk.Image.new_from_icon_name("go-previous",
                                         Gtk.IconSize.BUTTON))
        self.internet_toggle.hide()

    def on_keylist_mapped(self, keylistwidget):
        log.debug("Keylist becomes visible!")
        self.header_button.set_image(
            Gtk.Image.new_from_icon_name("view-refresh",
            Gtk.IconSize.BUTTON))
        # We don't support refreshing for now.
        self.header_button.set_sensitive(False)
        self.internet_toggle.show()

    def on_send_stack_mapped(self, stack):
        log.debug("send stack becomes visible!")
        # Adjust the top bar buttons
        self.on_send_stack_switch(stack)

    def on_keypresent_mapped(self, kpw):
        log.debug("keypresent becomes visible!")
        self.header_button.set_sensitive(True)
        self.header_button.set_image(
            Gtk.Image.new_from_icon_name("go-previous",
            Gtk.IconSize.BUTTON))
        self.internet_toggle.hide()

    def on_send_header_button_clicked(self, button, *args):
        # Here we assume that there is only two places where
        # we could have possibly pressed this button, i.e.
        # from the keypresentwidget or the result page
        log.debug("Send Headerbutton %r clicked! %r", button, args)
        current = self.send_app.stack.get_visible_child()
        klw = self.send_app.klw
        kpw = self.send_app.kpw
        # If we are in the keypresentwidget
        if current == kpw:
            self.send_stack.set_visible_child(klw)
            self.send_app.deactivate()
        # Else we are in the result page
        else:
            self.send_stack.remove(current)
            self.send_app.set_saved_child_visible()
            self.send_app.on_key_activated(None, self.send_app.key)

    def on_header_button_clicked(self, button, *args):
        log.debug("Headerbutton %r clicked! %r", button, args)
        return self.on_send_header_button_clicked(button, *args)

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    app = App(application_id="org.gnome.Keysign.Send")
    try:
        GLib.unix_signal_add_full(GLib.PRIORITY_HIGH, signal.SIGINT,
                                  lambda *args: reactor.callFromThread(reactor.stop), None)
    except AttributeError:
        pass
    reactor.registerGApplication(app)
    reactor.run()



#!/usr/bin/env python
#    Copyright 2015 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from .util import sign_keydata_and_send

if sys.version_info.major < 3:
    input = raw_input


def main():
    import argparse
    parser = argparse.ArgumentParser(description=""
        "Certify UIDs of given OpenPGP certificates (aka sign OpenPGP keys) "
        " from a file.  "
        "The program will open each file, extract exactly one "
        "OpenPGP certificate, sign each UID separately, encrypt and send "
        "each signed UID using xdg-email.")
    parser.add_argument('-v', '--verbose', action='count', default=0,
        help="Increase detail of logging")
    parser.add_argument("file", nargs='+', type=argparse.FileType('rb'),
        help="File containing OpenPGP keys")
    parser.add_argument('--sign-and-send-all-uids', action='store_true',
        help="Attach certifications for email-less UIDs to the emails dispatched for UIDs with emails.")
    args = parser.parse_args()

    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = log_levels[min(len(log_levels)-1, args.verbose)]
    logging.basicConfig(level=log_level)

    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)
    for fhandle in args.file:
        data = fhandle.read()
        log.info("Calling %r to sign %s", sign_keydata_and_send, fhandle.name)
        tmpfiles = list(sign_keydata_and_send(keydata=data, send_all_uids=args.sign_and_send_all_uids))
    print("Finished signing. " +
             "We're only waiting for the signature " +
             "files to be picked up. " +
             "Press any key to quit the application.")
    input()


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    sys.exit(main())



#!/usr/bin/env python
#    Copyright 2014 Tobias Mueller <muelli@cryptobitch.de>
#    Copyright 2014 Andrei Macavei <andrei.macavei89@gmail.com>
#    Copyright 2014 Srdjan Grubor <sgnn7@sgnn7.org>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

'''This is an exercise to see how we can combine Python threads
with the Gtk mainloop
'''

import logging
import os
import sys

from threading import Thread
from gi.repository import GLib
from gi.repository import Gtk
from dbus.mainloop.glib import DBusGMainLoop

if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')

from . import Keyserver

class ServerWindow(Gtk.Window):
    def __init__(self):
        self.log = logging.getLogger(__name__)

        Gtk.Window.__init__(self, title="Gtk and Python threads")
        self.set_border_width(10)

        self.connect("delete-event", Gtk.main_quit)

        hBox = Gtk.HBox()
        self.button = Gtk.ToggleButton('Start')
        hBox.pack_start(self.button, False, False, 0)
        self.add(hBox)

        self.button.connect('toggled', self.on_button_toggled)

        #GLib.idle_add(self.setup_server)

    def on_button_toggled(self, button):
        self.log.debug('toggled button')
        if button.get_active():
            self.log.debug("I am being switched on")
            self.setup_server()
        else:
            self.log.debug("I am being switched off")
            self.stop_server()

    def setup_server(self):
        self.log.info('Serving now')
        self.log.debug('About to call %r', Keyserver.ServeKeyThread)
        self.keyserver = Keyserver.ServeKeyThread(b'Keydata', 'fingerprint')
        self.log.info('Starting thread %r', self.keyserver)
        self.keyserver.start()
        self.log.info('Finsihed serving')

        return False

    def stop_server(self):
        self.keyserver.shutdown()

def main(args):
    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)
    w = ServerWindow()
    w.show_all()
    log.debug('Starting main')

    DBusGMainLoop(set_as_default = True)
    Gtk.main()

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    # From http://stackoverflow.com/a/16486080/2015768
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    sys.exit(main(sys.argv))



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys

from requests.exceptions import ConnectionError

from gi.repository import GObject, GLib


if  __name__ == "__main__" and __package__ is None:
    logging.getLogger().error("You seem to be trying to execute " +
                              "this script directly which is discouraged. " +
                              "Try python -m instead.")
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    os.sys.path.insert(0, parent_dir)
    os.sys.path.insert(0, os.path.join(parent_dir, 'monkeysign'))
    import keysign
    #mod = __import__('keysign')
    #sys.modules["keysign"] = mod
    __package__ = str('keysign')


from .util import strip_fingerprint, download_key_http, parse_barcode

try:
    from .gpgmeh import fingerprint_from_keydata
except ImportError:
    # FIXME: Remove this conditional
    from .gpgmeh import fingerprint_for_key as fingerprint_from_keydata

from .network.AvahiBrowser import AvahiBrowser
from .util import mac_verify

log = logging.getLogger(__name__)

class AvahiKeysignDiscovery(GObject.GObject):
    "A client discovery using Avahi"

    __gsignals__ = {
        # Gets emitted whenever a new server has been found or has been removed.
        # Is also emitted shortly after an object has been created.
        str("list-changed"): (GObject.SignalFlags.RUN_LAST, None, (int,)),
    }

    def __init__(self, *args, **kwargs):
        super(AvahiKeysignDiscovery, self).__init__(*args, **kwargs)
        self.log = logging.getLogger(__name__)
        # We should probably try to put this constant in a more central place
        avahi_service_type = '_gnome-keysign._tcp'
        self.avahi_browser = AvahiBrowser(service=avahi_service_type)
        self.avahi_browser.connect('new_service', self.on_new_service)
        self.avahi_browser.connect('remove_service', self.on_remove_service)
        self.discovered_services = []
        # It seems we cannot emit directly...
        GLib.idle_add(lambda: self.emit("list-changed",
            len(self.discovered_services)))

    def on_new_service(self, browser, name, address, port, txt_dict):
        published_fpr = txt_dict.get('fingerprint', None)
        self.log.info("discovered something: %s %s:%i:%s",
                      name, address, port, published_fpr)
        if not address.startswith('fe80::'):
            # We intend to ignore IPv6 link local addresses, because it seems
            # that you cannot just connect to that address without also
            # knowing which NIC the address belongs to.
            # http://serverfault.com/a/794967
            # FIXME: Use something more sane like attr.s instead of the tuple
            self.discovered_services += ((name, address, port, published_fpr), )
            self.emit("list-changed", len(self.discovered_services))

    def on_remove_service(self, browser, service_type, name):
        '''Handler for the on_remove signal from AvahiBrowser

        Removes a service from the internal list by calling
        remove_discovered_service.
        '''
        self.log.info("Received a remove signal, let's check; %s:%s",
                      service_type, name)
        self.remove_discovered_service(name)

    def remove_discovered_service(self, name):
        '''Removes server-side clients from discovered_services list
        when the server name with fpr is a match.'''
        for client in self.discovered_services:
            if client[0] == name:
                self.discovered_services.remove(client)
                self.emit("list-changed", len(self.discovered_services))
        self.log.info("Clients currently in list '%s'",
                      self.discovered_services)

    def find_key(self, userdata):
        "Returns the key if it thinks it found one..."
        self.log.info("Trying to find key with %r", userdata)
        parsed = parse_barcode(userdata)
        cleaned = strip_fingerprint(parsed["fingerprint"])
        downloaded_key = None
        # FIXME: Replace with attr.ib
        for (name, address, port, fpr) in self.discovered_services:
            if cleaned == fpr:
                # This is blocking :-/
                try:
                    downloaded_key = download_key_http(address, port)
                    if fingerprint_from_keydata(downloaded_key) != cleaned:
                        continue
                except ConnectionError:
                    self.log.exception("Error downloading from %r:%r",
                                  address, port)
        return downloaded_key

class AvahiKeysignDiscoveryWithMac(AvahiKeysignDiscovery):
    def find_key(self, userdata):
        "Returns the key if it thinks it found one which also matched the MAC"
        key = super(AvahiKeysignDiscoveryWithMac, self).find_key(userdata)
        if key:
            # For now, we cannot assume that a MAC exists, simply because
            # currently the MAC is only transferred via the barcode.
            # The user, however, might as well enter the fingerprint
            # manually.  Unless we stop allowing that, we won't have a MAC.
            mac = parse_barcode(userdata).get("MAC", [None])[0]
            if mac is None:
                # This is the ugly shortcut which exists for legacy reasons
                verified_key = key
            else:
                mac_key = fingerprint_from_keydata(key)
                verified = mac_verify(mac_key.encode('ascii'), key, mac)
                if verified:
                    verified_key = key
                else:
                    self.log.info("MAC validation failed: %r", verified)
                    verified_key = None
        else:
            verified_key = None

        return verified_key

def main(args):
    log = logging.getLogger(__name__)
    log.debug('Running main with args: %s', args)
    if not args:
        raise ValueError("You must provide an argument to identify the key")

    loop = GObject.MainLoop()

    arg = args[0]
    # FIXME: Enable parameter
    timeout = 5
    GObject.timeout_add_seconds(timeout, lambda: loop.quit())

    discover = AvahiKeysignDiscovery()
    # We quickly attach the found to the object to maintain state
    discover.found_key = None
    def find_key():
        keydata = discover.find_key(arg)
        if keydata:
            log.info("Found %d key bytes", len(keydata))
            discover.found_key = keydata
            print (keydata)
            loop.quit()
        return not keydata

    discover.avahi_browser.connect('new_service', lambda *args: find_key())
    # Instead of using this implementation detail for getting the notification,
    # it would be possible to repeatedly call find_key.
    # GObject.timeout_add(25, lambda: find_key() and False)
    # GObject.timeout_add(500, find_key)
    loop.run()
    if not discover.found_key:
        log.error("No Key found for %r!!1", arg)

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG,
            format='%(name)s (%(levelname)s): %(message)s')
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
#    Copyright 2016 Tobias Mueller <muelli@cryptobitch.de>
#
#    This file is part of GNOME Keysign.
#
#    GNOME Keysign is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNOME Keysign is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GNOME Keysign.  If not, see <http://www.gnu.org/licenses/>.
import base64
import logging
import os  # The SigningKeyring uses os.symlink for the agent
from subprocess import check_output
import sys
from tempfile import mkdtemp
import platform

import dbus
import gpg
from gpg.constants import PROTOCOL_OpenPGP, IMPORT_NEW, IMPORT_SIG
from gpg.errors import GPGMEError


from .gpgkey import Key, UID

texttype = unicode if sys.version_info.major < 3 else str

log = logging.getLogger(__name__)


#####
## INTERNAL API
##

class GPGRuntimeError(RuntimeError):
    pass

class NoSecretKeysError(GPGRuntimeError):
    def __init__(self, message, homedir=None, all_keys=None):
        super(NoSecretKeysError, self).__init__(message)
        self.homedir = homedir
        self.all_keys = all_keys or []
        
    def __str__(self):
        return "{msg} (Homedir: {homedir}, Keys found: {keys})".format(
            msg=super(NoSecretKeysError, self).__str__(),
            homedir=self.homedir,
            keys=len(self.all_keys)
        )

class GenEdit:
    _ignored_status = (gpg.constants.STATUS_EOF,
                       gpg.constants.STATUS_GOT_IT,
                       gpg.constants.STATUS_NEED_PASSPHRASE,
                       gpg.constants.STATUS_GOOD_PASSPHRASE,
                       gpg.constants.STATUS_BAD_PASSPHRASE,
                       gpg.constants.STATUS_USERID_HINT,
                       gpg.constants.STATUS_SIGEXPIRED,
                       gpg.constants.STATUS_KEYEXPIRED,
                       gpg.constants.STATUS_PROGRESS,
                       gpg.constants.STATUS_KEY_CREATED,
                       gpg.constants.STATUS_ALREADY_SIGNED,
                       gpg.constants.STATUS_KEY_CONSIDERED,
                       gpg.constants.STATUS_CARDCTRL)

    def __init__(self, generator):
        generator.send(None)
        self.generator = generator
        self.last_sink_index = 0

    def edit_cb(self, status, args, sink=None):
        if status in self._ignored_status:
                logging.info("Returning None for %r %r", status, args)
                return
        if not status:
            logging.info("Closing for %r", status)
            self.generator.close()
            return
        # 0 is os.SEEK_SET
        if sink:
            # os.SEEK_CUR = 1
            current = sink.seek(0, 1)
            sink.seek(self.last_sink_index, 0)
            sinkdata = sink.read(current)
            self.last_sink_index = current
        else:
            sinkdata = None
        log.info("edit_cb: %r %r '%s'", status, args, sinkdata)
        data = self.generator.send((status, args)) #, sinkdata))
        log.info("edit_cb data: %r", data)
        return texttype(data)

def del_uids(uids):
    status, arg = yield None
    log.info("status args: %r %r", status, arg)
    #log.info("sinkdata: %s", sinkdata)
    #uids = [l for l in sinkdata.splitlines() if l.startswith('uid:')]
    #log.info("UIDs: %s", uids)
    status, arg = yield "list"
    log.info("status args: %r %r", status, arg)
    for uid in uids:
        status, arg = yield "uid %d" % uid
        log.info("status args: %r %r", status, arg)
    if uids:
        status, arg = yield "deluid"
        log.info("status args: %r %r", status, arg)
        assert status == gpg.constants.STATUS_GET_BOOL, "%r %r" % (status, arg)
        assert arg == 'keyedit.remove.uid.okay'
        status, arg = yield "Y"
        log.info("status args: %r %r", status, arg)
    yield 'save'


def sign_key(uid=0, sign_cmd=u"sign", expire=False, check=3,
             error_cb=None):
    log.info("Signing key uid %r", uid)
    status, prompt = yield None
    if status == u"status_code_lost": # Yeah, such a constant does not exist *sigh*
        # We are here, because the agent on the host is too old for
        # what the guest, e.g. the flatpaked app, expects.
        # Let's hope we can just ignore that.
        log.info("Agent on the host might be too old: %s %s",
            status, prompt)
        status, prompt = yield None

    assert status == gpg.constants.STATUS_GET_LINE, "Expected status to be GET_LINE, but is %r" % status
    assert prompt == u"keyedit.prompt"

    status, prompt = yield u"uid %d" % uid
    # We ignore GOT_IT...
    # assert status == gpg.constants.STATUS_GOT_IT

    #status, prompt = yield None
    assert status == gpg.constants.STATUS_GET_LINE

    status, prompt = yield sign_cmd
    # We ignore GOT_IT...
    # assert status == gpg.constants.STATUS_GOT_IT

    while prompt != 'keyedit.prompt':
        if prompt == 'keyedit.sign_all.okay':
            status, prompt = yield 'Y'
        elif prompt == 'sign_uid.expire':
            status, prompt = yield '%s' % ('Y' if expire else 'N')
        elif prompt == 'sign_uid.class':
            status, prompt = yield '%d' % check
        elif prompt == 'sign_uid.okay':
            status, prompt = yield 'Y'
        elif status == gpg.constants.STATUS_INV_SGNR:
            # seems to happen if you have an expired
            # (or otherwise unusable) signing key.
            # The CONSIDERED line should have been issued
            # with details.
            # We don't maintain that state at the moment which is
            # a bit unfortunate as we cannot properly detect
            # when we have no usable key at all rather than
            # one key being expired.
            log.warn("INV_SGNR: %r", prompt)
            status, prompt = yield None
        elif status == gpg.constants.STATUS_PINENTRY_LAUNCHED:
            status, prompt = yield None
        elif status == gpg.constants.STATUS_GOT_IT:
            status, prompt = yield None
        elif status == gpg.constants.STATUS_ALREADY_SIGNED:
            status, prompt = yield u'Y'
        elif status == gpg.constants.STATUS_ERROR:
            if error_cb:
                error_cb(prompt)
            else:
                raise GPGRuntimeError("Error signing key: %s" % prompt)
            status, prompt = yield None
        else:
            raise AssertionError("Unexpected state %r %r" % (status, prompt))

    yield u"save"




def UIDExport(keydata, uid_i):
    """Export only the UID of a key.
    Unfortunately, GnuPG does not provide smth like
    --export-uid-only in order to obtain a UID and its
    signatures."""
    log.debug("Deletion of UID %r from %r", uid_i, keydata)
    if not uid_i >= 1:
        log.debug("Raising because uid: %r", uid_i)
        raise ValueError("Expected UID to be >= 1, but is %r", uid_i)
    ctx = TempContext()
    ctx.op_import(keydata)
    result = ctx.op_import_result()
    if result.considered != 1 or result.imported != 1:
        raise ValueError("Expected exactly one key in keydata. %r" % result)
    else:
        assert len(result.imports) == 1
        fpr = result.imports[0].fpr
        key = ctx.get_key(fpr)
        uids_to_remove = {i for i in range(1, len(key.uids)+1)}
        uids_to_remove.remove(uid_i)
        if uids_to_remove:
            sink = gpg.Data()
            ctx.interact(key,
                GenEdit(del_uids(uids_to_remove)).edit_cb,
                fnc_value=sink, sink=sink)
            sink.seek(0, 0)
            log.debug("Data after UIDExport: %s", sink.read())
        uid_data = gpg.Data()
        ctx.op_export_keys([key], 0, uid_data)
        uid_data.seek(0, 0)
        uid_bytes = uid_data.read()
        log.debug("UID %r: %r", uid_i, uid_bytes)
        return uid_bytes

def export_uids(keydata):
    """Export each valid and non-revoked UID of a key

    Returns: An iterator over a tuple of the UID, i.e. the string and
    the bytes making the OpenPGP with that UID only.
    """
    ctx = TempContext()
    ctx.op_import(keydata)
    result = ctx.op_import_result()
    log.debug("ExportUIDs: Imported %r", result)
    if result.considered != 1 or result.imported != 1:
        raise ValueError("Expected exactly one key in keydata. %r" % result)
    else:
        assert len(result.imports) == 1
        fpr = result.imports[0].fpr
        key = ctx.get_key(fpr)
        for i, uid in enumerate(key.uids, start=1):
            log.info("Potentially deleting UID %d: %r", i, uid)
            if not uid.invalid and not uid.revoked:
                uid_data = UIDExport(keydata, i)
                yield (uid.uid, uid_data)





def is_usable(key):
    unusable =    key.invalid or key.disabled \
               or key.expired or key.revoked
    log.debug('Key %s is invalid: %s (i:%s, d:%s, e:%s, r:%s)', key, unusable,
        key.invalid, key.disabled, key.expired, key.revoked)
    return not unusable

def filter_usable_keys(keys):
    usable_keys = [Key.from_gpgme(key) for key in keys if is_usable(key)]
    log.debug('Identified usable keys: %s', usable_keys)
    return usable_keys



class DirectoryContext(gpg.Context):
    def __init__(self, homedir):
        super(DirectoryContext, self).__init__()
        self.set_engine_info(PROTOCOL_OpenPGP, None, homedir)
        self.homedir = homedir

class TempContext(DirectoryContext):
    def __init__(self):
        self.homedir = mkdtemp(suffix="-gnome-keysign")
        super(TempContext, self).__init__(homedir=self.homedir)

    def __del__(self):
        try:
            # shutil.rmtree(self.homedir, ignore_errors=True)
            pass
        except:
            log.exception("During cleanup of %r", self.homedir)

def get_agent_socket_path_for_homedir(homedir):
    homedir_cmd = ["--homedir", homedir] if homedir else []
    cmd = ["gpgconf"] + homedir_cmd + \
          ["--list-dirs", "agent-socket"]
    path = check_output(cmd).strip()
    log.info("Path for %r: %r", homedir, path)
    return path


class TempContextWithAgent(TempContext):
    def __init__(self, oldctx):
        super(TempContextWithAgent, self).__init__()
        homedir = self.homedir
        log.info("new homedir: %r", homedir)
        assert (len(list(self.keylist())) == 0)
        assert (len(list(self.keylist(secret=True))) == 0)


        old_homedir = oldctx.engine_info.home_dir if oldctx else None

        log.info("Old homedir: %r", old_homedir)
        old_agent_path = get_agent_socket_path_for_homedir(old_homedir)
        new_agent_path = get_agent_socket_path_for_homedir(homedir)
        os.symlink(old_agent_path, new_agent_path)

        assert len(list(self.keylist())) == 0
        assert len(list(self.keylist(secret=True))) == 0

        secret_keys = list(oldctx.keylist(secret=True))
        log.info("old secret keys: %r", secret_keys)
        for key in secret_keys:
            log.debug("Making %r known in new ctx", key)
            def export_key(fpr):
                # FIXME: The Context should really be able to export()
                public_key = gpg.Data()
                oldctx.op_export(fpr, 0, public_key)
                public_key.seek(0, os.SEEK_SET)
                return public_key
            keydata = export_key(key.subkeys[0].fpr)
            self.op_import(keydata)
            result = self.op_import_result()
            # Hrm. Only gpgme>=1.9 has a repr for the result, I think
            log.debug("Import result: %r", result)
            log.debug("Import result imports: %r", result.imports)
            log.debug("Import result considered: %r", result.considered)
            assert len(result.imports) >= 1
            i = result.imports[0]
            # 0 is success, I guess.
            assert i.result == 0
            log.debug("Import result i result status: %r %r %r", i.result, i.status, i.fpr)
            log.debug("Import result GPGME_IMPORT_NEW: %r", i.status & gpg.constants.IMPORT_NEW)


        assert len(list(self.keylist())) == len(secret_keys)
        log.info("new secret keys: %r", list(self.keylist(secret=True)))
        assert len(secret_keys) == len(list(self.keylist(secret=True)))



def import_signature_dbus(signature):
    "Imports a TPK into the user's keyring by using Seahorse's DBus API"
    name = "org.gnome.seahorse"
    path = "/org/gnome/seahorse/keys"
    bus = dbus.SessionBus()
    result = []

    proxy = bus.get_object(name, path)
    iface = "org.gnome.seahorse.KeyService"
    gpg_iface = dbus.Interface(proxy, iface)
    payload = base64.b64encode(signature).decode('latin-1')
    payload = '\n'.join(payload[i:(i + 64)] for i in range(0, len(payload), 64))
    payload = "-----BEGIN PGP PUBLIC KEY BLOCK-----\n\n" + payload + "\n-----END PGP PUBLIC KEY BLOCK-----"
    result = gpg_iface.ImportKeys("openpgp", payload)
    log.debug("Importing via DBus: %r", result)

    return result

def import_signature_gpgme(signature, homedir=None):
    "Imports an OpenPGP TPK into a keyring via GPGME"
    ctx = DirectoryContext(homedir)
    ctx.op_import(signature)
    result = ctx.op_import_result()
    if len(result.imports) < 1:
        raise GPGMEError

    return result



##
## END OF INTERNAL API
#####




def get_signatures_for_uids_on_key(key, homedir=None):
    """It seems to be a bit hard to get a key with its signatures,
    so this is a small helper function"""
    # esp. get_key does not take a SIGS argument.
    # What happens if keylist returns multiple keys, e.g. because there
    # is another key with a UID named as the fpr?  How can I make sure I
    # get the signatures of any given key?
    
    # *sigh* gpgme is killing me. With gpgme 1.8 we have to
    # set_keylist_mode before we can call keylist.  With gpgme 1.9
    # keylist takes a mode argument and overrides whatever has been
    # set before.  In order to come with something compatible with both
    # 1.8 and 1.9 we have to set_keylist_mode and NOT call ctx.keylist
    # but rather the bare op_keylist_all.  In 1.8 that requires two
    # arguments.
    ctx = DirectoryContext(homedir)
    mode = gpg.constants.keylist.mode.LOCAL | gpg.constants.keylist.mode.SIGS
    secret = False
    ctx.set_keylist_mode(mode)
    keys = list(ctx.op_keylist_all(key.fpr, secret))
    # With gpgme 1.9 we can simply do:
    # keys = list(ctx.keylist(key.fpr), mode=mode)
    assert len(keys) == 1
    uid_sigs = {uid.uid: {s.keyid for s in uid.signatures} for uid in keys[0].uids}
    log.info("Signatures: %r", uid_sigs)
    return uid_sigs



def openpgpkey_from_data(keydata):
    c = TempContext()
    c.op_import(gpg.Data(keydata))
    result = c.op_import_result()
    log.debug("Import Result: %s", result)
    if result.imported != 1:
        raise ValueError("Keydata did not contain exactly one key, but %r" %
            result.imported)
    else:
        imported = result.imports
        import_ = imported[0]
        fpr = import_.fpr
        key = c.get_key(fpr)
        return Key.from_gpgme(key)



def get_public_key_data(fpr, homedir=None):
    c = DirectoryContext(homedir)
    c.armor = True
    sink = gpg.Data()
    # FIXME: There will probably be an export() function
    c.op_export(fpr, 0, sink)
    sink.seek(0, os.SEEK_SET)
    keydata = sink.read()
    log.debug("Exported %r: %r", fpr, keydata)
    if not keydata:
        s = "No data to export for {} (in {})".format(fpr, homedir)
        raise ValueError(s)
    return keydata



def fingerprint_from_keydata(keydata):
    '''Returns the OpenPGP Fingerprint for a given key'''
    openpgpkey = openpgpkey_from_data(keydata)
    return openpgpkey.fingerprint

def get_usable_keys_from_context(ctx, pattern="", secret=False):
    keys = [Key.from_gpgme(key)
            for key in ctx.keylist(pattern=pattern, secret=secret)
            if is_usable(key)
                and     # We filter "offline" keys.
                        # If secret=False then it passes.
                        # If secret=True then we require the primary subkey
                        # to be "secret". It is false if the key is offline.
                (not secret or key.subkeys[0].secret)
            ]
    return keys

def get_usable_keys(pattern="", homedir=None):
    '''Uses get_keys on the keyring and filters for
    non revoked, expired, disabled, or invalid keys'''
    log.debug('Retrieving keys for %s, %s', pattern, homedir)
    ctx = DirectoryContext(homedir=homedir)
    return get_usable_keys_from_context(ctx,
    	pattern=pattern, secret=False)

def get_usable_secret_keys(pattern="", homedir=None):
    '''Returns all secret keys which can be used to sign a key'''
    ctx = DirectoryContext(homedir=homedir)
    return get_usable_keys_from_context(ctx,
    	pattern=pattern, secret=True)




def minimise_key(keydata):
    "Returns the public key exported under the MINIMAL mode"
    ctx = TempContext()
    ctx.op_import(keydata)
    result = ctx.op_import_result()
    if result.considered != 1 and result.imported != 1:
        raise ValueError("Expected to load exactly one key. %r", result)
    else:
        imports = [i for i in result.imports
                   if i.status == gpg.constants.IMPORT_NEW]
        log.debug("Import %r", result)
        assert len(imports) == 1
        fpr = result.imports[0].fpr
        key = ctx.get_key(fpr)
        sink = gpg.Data()
        ctx.op_export_keys([key], gpg.constants.EXPORT_MODE_MINIMAL, sink)
        sink.seek(0, 0)
        minimised_key = sink.read()
        return minimised_key

def local_sign_keydata(keydata, expires_in=60*60*24*1, error_cb=None, homedir=None):
    """Produces non-exportable and expiring signatures.
    This can be useful if we want to enable the user to send an email to 
    the other party right away, without waiting for the protocol to have 
    completed. By letting the signature expire, we limit the time in 
    which a wrongly signed key is harmful. This is a challenge for thx
    UX, because sending emails will stop working pretty much out of 
    the blue. But it can hardly be any worse than it is now.
    And the app ought to inform the user about the fact that it's only 
    ephemeral.

    Returns: nothing
    """
    ctx = DirectoryContext(homedir)

    tmpctx = TempContext()
    available_secret_keys = [key for key in ctx.keylist(secret=True)
        if not key.disabled or key.revoked or key.invalid or key.expired]
    log.debug('Setting available sec keys to: %r', available_secret_keys)
    ctx.signers = available_secret_keys

    tmpctx.op_import(keydata)
    result = tmpctx.op_import_result()
    if result.considered != 1 and result.imported != 1:
        raise ValueError("Expected to load exactly one key. %r", result)
    else:
        imports = result.imports
        assert len(imports) == 1
        fpr = result.imports[0].fpr

        ctx.op_import(keydata)
        key = ctx.get_key(fpr)
        # We need to sign in the regular context, because gpgme does not
        # export local signatures from a keyring.
        ctx.key_sign(key, local=True, expires_in=expires_in)
        # Unfortunately, key_sign does not report back how many
        # signatures were produced (or not produced...)
        # It may raise an error, but I have yet to see that it does...
        log.info("Locally signed key %s with an expiry in %d seconds", fpr, expires_in)


def sign_keydata_and_encrypt(keydata, error_cb=None, homedir=None):
    oldctx = DirectoryContext(homedir)
    ctx = TempContextWithAgent(oldctx)
    # We're trying to sign with all available secret keys
    all_secret_keys = list(ctx.keylist(secret=True))
    available_secret_keys = [key for key in all_secret_keys
        if not (key.disabled or key.revoked or key.invalid or key.expired)]
    log.debug('Setting available sec keys to (%d): %r',
        len(available_secret_keys), available_secret_keys)
    if not available_secret_keys:
        raise NoSecretKeysError(
            "No secret keys available to sign with.",
            homedir=homedir,
            all_keys=all_secret_keys
        )
    ctx.signers = available_secret_keys

    ctx.op_import(minimise_key(keydata))
    result = ctx.op_import_result()
    if result.considered != 1 and result.imported != 1:
        raise ValueError("Expected to load exactly one key. %r", result)
    else:
        imports = result.imports
        assert len(imports) == 1
        fpr = result.imports[0].fpr
        key = ctx.get_key(fpr)
        sink = gpg.Data()
        # There is op_keysign, but it's only available with gpg 2.1.12
        ctx.interact(key, GenEdit(sign_key(error_cb=error_cb)).edit_cb, sink=sink)
        sink.seek(0, 0)
        log.debug("Sink after signing: %r", sink.read())

        signed_sink = gpg.Data()
        ctx.set_keylist_mode(gpg.constants.KEYLIST_MODE_SIGS)
        ctx.armor = True
        ctx.op_export_keys([key], 0, signed_sink)
        signed_sink.seek(0, 0)
        signed_keydata = signed_sink.read()
        log.debug("Signed Key: %s", signed_keydata)
        # Do I have to re-get the key to make the signatures known?
        key = ctx.get_key(fpr)

        for i, uid in enumerate(key.uids, start=1):
            if uid.revoked or uid.invalid:
                continue
            else:
                log.debug("The UID %s has %d signatures",
                    uid, len(uid.signatures))
                if len(uid.signatures) < 2:
                    log.error("We seem to not have produced signatures correctly. "
                        "%s has less than 2 signatures: %s",
                        uid, uid.signatures,
                    )

                uid_data = UIDExport(signed_keydata, i)
                log.debug("Data for uid %d: %r, sigs: %r %r", i, uid, uid.signatures, uid_data)

                ciphertext, _, _ = ctx.encrypt(plaintext=uid_data,
                                               recipients=[key],
                                               # We probably have to set owner trust
                                               # in order for it to work out of the box
                                               always_trust=True,
                                               sign=False)
                yield (UID.from_gpgme(uid), ciphertext, uid_data)


class NoNewSignatures(GPGMEError):
    "We couldn't find a new certification, so the certification is already known"
    def __init__(self, signature, import_result):
        self.signature = signature
        self.import_result = import_result
        super().__init__()
class NewRevocations(GPGMEError):
    pass
class NewSubkey(GPGMEError):
    pass

def decrypt_signature(encrypted_sig, homedir=None):
    """
    Takes an encrypted signture, tries to decrypt it, and returns the
    decrypted signature if it is does indeed contain a certification only
    """
    ctx = DirectoryContext(homedir)

    # Check if we are really importing a signature
    temp_ctx = TempContextWithAgent(ctx)
    signature = temp_ctx.decrypt(encrypted_sig)
    log.debug("signature decryption result: %r", signature)
    decrypted_sig = signature[0]
    temp_ctx.op_import(decrypted_sig)
    result = temp_ctx.op_import_result()
    log.debug("signature import result: %r", result)

    if result.imported != 0:
        log.warning("Trying to import a new key instead of a signature!")
        raise GPGMEError

    if result.new_signatures == 0:
        raise NoNewSignatures(signature=decrypted_sig, import_result=result)
    if result.new_revocations != 0:
        raise NewRevocations()
    if result.new_sub_keys != 0:
        raise NewSubkey()

    return decrypted_sig

def decrypt_and_import_signature(encrypted_sig, homedir=None):
    signature = decrypt_signature(encrypted_sig, homedir=homedir)
    import_signature(signature)
    return signature




class ImportNewCertificationError(GPGMEError):
    "The import of a TPK failed, probably due to containing a new certificate rather than new 'signatures'"


def import_signature(signature, homedir=None):
    """
    Imports an OpenPGP TPK to the local keyring

    The purpose is to import a certification (hence the name of the function)
    but it is in fact agnostic about what the TPK contains.

    This function will try to import the TPK via DBus first and,
    if that failed, resort to using gpgme directly.
    """
    result = []

    ctx = TempContextWithAgent(DirectoryContext(homedir=homedir))
    ctx.op_import(signature)
    res = ctx.op_import_result()
    log.debug("ImportSignature: Testing for new certificate: %r", res)
    imports = res.imports
    if len(imports) != 1:
        log.error("We expected to import only one certificate, "
                  "but it seems we have %d", len(import_))
        raise ImportNewCertificationError
    else:
        import_ = imports[0]
        if import_.status & gpg.constants.IMPORT_NEW:
            log.error("We did not expect to import a *new* certificate, "
                      "but this seems to be new: %r", import_)
            raise ImportNewCertificationError
        else:
            assert import_.status & gpg.constants.IMPORT_SIG

            if not homedir:
                # If a homedir is requested, we have to use the gpgme API, because we cannot specify a GnuPG keyring via DBus
                try:
                    # Try Seahorse DBus
                    result = import_signature_dbus(signature)
                except dbus.exceptions.DBusException:
                    log.debug("Seahorse DBus is not available")

            # If Seahorse failed we try op_import
            if len(result) < 1:
                result = import_signature_gpgme(signature, homedir=homedir)

        return result
